from io import StringIO
import networkx as nx
from sklearn.tree import export_graphviz
import pydotplus

import plotly.graph_objs as go
from plotly.offline import iplot

from palantiri.BasePlotHandlers import PlotHandler


class GraphPlotHandler(PlotHandler):
    """
    The graph plot handler - handles all the graph related plots.
    """

    def __init__(self, graph, node_data_key=None, **params):
        """
        Initialization function
        :param graph: the graph to be plotted.
        :param node_data_key: the location of the relevant data within each node.
        :param params: other params
        """
        self._graph = graph
        self.node_data_key = node_data_key
        self.graph_figure = None

        super(GraphPlotHandler, self).__init__(**params)

    @property
    def graph(self):
        """
        Getter for the graph
        :return: the graph
        """
        return self._graph

    @graph.setter
    def graph(self, graph):
        """
        Setter for the graph
        :param graph: the new graph.
        """
        self._graph = graph

    def _get_edge_trace(self, pos):
        """
        Constructing the edge trace.
        :param pos: the position of each node : {node_number:[x,y]}
        :return: a plot.ly trace object of the edges.
        """
        edge_trace = go.Scatter(
            x=[],
            y=[],
            line=dict(width=0.5, color='#888'),
            hoverinfo='none',
            mode='lines',
            showlegend=False)

        for edge in self.graph.edges():
            x0, y0 = pos[edge[0]]
            x1, y1 = pos[edge[1]]
            edge_trace['x'] += tuple([x0, x1, None])
            edge_trace['y'] += tuple([y0, y1, None])

        return edge_trace

    def _get_node_trace(self, pos):
        """
        Constructing the node trace
        :param pos: the position of each node: {node_number:[x,y]}.
        :return: a plot.ly trace object og the nodes
        """

        node_trace = go.Scatter(
            x=[],
            y=[],
            text=[],
            mode='markers',
            hoverinfo='text',
            showlegend=False,
            marker=dict(
                showscale=False,
                colorscale='YlGnBu',
                reversescale=True,
                color=[],
                size=17,
                line=dict(width=2)))

        for node, node_data in self.graph.nodes(data=True):
            x, y = pos[node]

            node_trace['x'] += tuple([x])
            node_trace['y'] += tuple([y])
            node_trace['text'] += tuple([node_data[self.node_data_key] if self.node_data_key else node])
            if 'color' in node_data.keys():
                node_trace['marker']['color'] += tuple([hash(node_data['color']) % 256])
            else:
                node_trace['marker']['color'] += tuple([hash(node_data[self.node_data_key]) % 256])
        return node_trace

    def build_graph_figure(self, figure_layout=go.Layout(), pos=None):
        """
        Building the graph plot figure.
        :param figure_layout: a plot.ly layout object.
        :param pos: the position of each node - if none: spring layout is used for position - {node_number:[x,y]}.
        """

        if not pos:
            pos = nx.spring_layout(self._graph)

        data = [self._get_edge_trace(pos), self._get_node_trace(pos)]

        self.graph_figure = go.Figure(data=data, layout=figure_layout)

    def plot_graph(self, figure_layout=None):
        """
        Plotting the graph figure.
        :param figure_layout: a plot.ly layout object.
        """

        if not figure_layout:
            figure_layout = go.Layout(title=dict(text='Graph Plot', x=0.5),
                                      xaxis=dict(
                                          autorange=True,
                                          showgrid=False,
                                          zeroline=False,
                                          showline=False,
                                          ticks='',
                                          showticklabels=False),
                                      yaxis=dict(
                                          autorange=True,
                                          showgrid=False,
                                          zeroline=False,
                                          showline=False,
                                          ticks='',
                                          showticklabels=False))

        if not self.graph_figure:
            self.build_graph_figure(figure_layout=figure_layout)

        else:
            self.graph_figure.layout = figure_layout

        iplot(self.graph_figure)


class DecisionTreePlotHandler(GraphPlotHandler):
    """
    The decision tree plot handler - handles all the decision tree related plots.
    """

    def __init__(self, decision_tree, feature_names, **params):
        """
        Initializing function for Decision tree plot handler
        :param decision_tree:  sklearn decision tree
        :param feature_names: the feature names
        :param params: other params
        :return: graph plot handler for the decision tree.
        """

        self.decision_tree = decision_tree
        self.feature_names = feature_names

        graph, pos = self._extract_graph_with_attributes_and_positions_from_tree()

        super().__init__(graph=graph, node_data_key='label', **params)

        self.build_graph_figure(pos=pos)

    def _extract_graph_with_attributes_and_positions_from_tree(self):
        """
        _extract_graph_with_attributes_and_positions_from_tree: extracts a graph representing the decision
        tree and the position of each node.
        :return: networkx graph an array
        """
        dot_data = StringIO()

        class_names = [str(name) for name in self.decision_tree.classes_]

        export_graphviz(self.decision_tree, out_file=dot_data, feature_names=self.feature_names,
                        class_names=class_names, filled=True)

        dot_graph = pydotplus.graph_from_dot_data(dot_data.getvalue())

        node_attributes = {node.obj_dict['name']: {
            'label': node.get_label().replace('\\n', '<br>').replace('"', ''),
            'color': node.get_label().replace('"', '').split('class')[-1].split(' ')[-1]}
            for node in dot_graph.get_nodes() if node.get_label()}

        edge_list = [edge.obj_dict['points'] for edge in dot_graph.get_edge_list()]

        nx_graph = nx.from_edgelist(edgelist=edge_list)

        pos = nx.drawing.nx_pydot.graphviz_layout(nx_graph, prog='dot')

        nx.set_node_attributes(nx_graph, node_attributes)

        return nx_graph, pos


class RandomForestPlotHandler(GraphPlotHandler):
    """
    The random forest plot handler - handles all the random forest related plots.
    """
    def __init__(self, random_forest, feature_names, **params):
        """
        Initializing function for random forest plot handler
        :param random_forest:  sklearn random forest
        :param feature_names: the feature names
        :param params: other params
        :return: graph plot handler for the decision tree.
        """
        self.feature_names = feature_names
        self.list_of_trees = random_forest.estimators_

        forest = self._build_forest_graph()

        super().__init__(graph=forest, node_data_key='label', **params)

    def _extract_graph_from_tree(self, decision_tree):
        """
        _extract_graph_from_tree: extracts a graph representing the decision tree.
        :return: networkx graph
        """

        dot_data = StringIO()

        class_names = [str(name) for name in decision_tree.classes_]

        export_graphviz(decision_tree, out_file=dot_data, feature_names=self.feature_names,
                        class_names=class_names, filled=True)

        dot_graph = pydotplus.graph_from_dot_data(dot_data.getvalue())

        edge_list = [edge.obj_dict['points'] for edge in dot_graph.get_edge_list()]

        nx_graph = nx.from_edgelist(edgelist=edge_list)

        node_attributes = {node.obj_dict['name']: {
            'label': node.get_label().replace('\\n', '<br>').replace('"', ''),
            'color': node.get_label().replace('"', '').split('class')[-1].split(' ')[-1]}
            for node in dot_graph.get_nodes() if node.get_label()}

        nx.set_node_attributes(nx_graph, node_attributes)

        return nx_graph

    def _build_forest_graph(self):
        """
        _build_forest_graph - building a networkx graph representing the forest
        :return: a networkx graph
        """
        forest = nx.DiGraph()
        forest.add_nodes_from(nodes_for_adding=['Root'])

        for i, tree_ in enumerate(self.list_of_trees):
            graph = self._extract_graph_from_tree(tree_)

            rename_map = {node: node + '_{0}'.format(i) for node in graph.nodes()}

            forest = nx.union(forest, nx.relabel_nodes(graph, rename_map))

            forest.add_edge('Root', '0_{0}'.format(i))

        rename_map = {node: j for j, node in enumerate(forest.nodes())}

        return nx.relabel_nodes(forest, rename_map)

    def build_graph_figure(self, figure_layout=go.Layout(), pos=None):
        """
        Building the graph plot figure.
        :param figure_layout: a plot.ly layout object.
        :param pos: place holder
        """

        structure = {node: "" for node in self.graph.nodes()}

        for edge in self.graph.edges():
            parent, label = edge
            structure[label] = parent

        hover_text = [node_data['label'] if 'label' in node_data else ''
                      for node, node_data in self.graph.nodes(data=True)]

        data = [go.Sunburst(
            labels=list(structure.keys()),
            parents=list(structure.values()),
            hovertext=hover_text,
            hoverinfo="text",
            textinfo='text',
            text=['' for _ in range(len(structure.values()))])]

        self.graph_figure = go.Figure(data=data, layout=figure_layout)

    def plot_graph(self, figure_layout=None):
        """
        Plotting the random forest graph figure.
        :param figure_layout: a plot.ly layout object.
        """

        if not figure_layout:
            figure_layout = go.Layout(title=dict(text='Random Forest Graph Plot', x=0.5),
                                      xaxis=dict(
                                          autorange=True,
                                          showgrid=False,
                                          zeroline=False,
                                          showline=False,
                                          ticks='',
                                          showticklabels=False),
                                      yaxis=dict(
                                          autorange=True,
                                          showgrid=False,
                                          zeroline=False,
                                          showline=False,
                                          ticks='',
                                          showticklabels=False))

        if not self.graph_figure:
            self.build_graph_figure(figure_layout=figure_layout)

        else:
            self.graph_figure.layout = figure_layout

        iplot(self.graph_figure)




import numpy as np
from sklearn.metrics import confusion_matrix, roc_curve, auc


import plotly.graph_objs as go
import plotly.figure_factory as ff
from plotly.offline import iplot

from palantiri.BasePlotHandlers import PlotHandler


class ClassifierPlotHandler(PlotHandler):
    """ Handles all the plots related of the chosen classifier. """

    def __init__(self, dataset, trained_classifier, **params):
        """
        Initialization function
        :param dataset: the dataset in a dict format with the following keys:
                        'data' - numpy array with all the data points.
                        'target' - the label of the corresponding data point.
                        'target_names' - the label name.

        :param trained_classifier: sklearn classifier (trained / fitted).
                                   In order to plot the ROC plot - the classifier should have the predict_proba ability.
        :param params: other params
        """

        self._dataset = dataset
        self._trained_classifier = trained_classifier

        self._n_classes = len(set(dataset['target']))

        if hasattr(self._dataset, 'target_names'):
            self.class_names = self._dataset['target_names']
        else:
            self.class_names = ['Class {0}'.format(i) for i in range(self.n_classes)]

        # Score of the predicted target store.
        if hasattr(self._trained_classifier, 'predict_proba'):
            self._predicted_target_score = self._trained_classifier.predict_proba(self._dataset['data'])
        else:
            self._predicted_target_score = None

        self._confusion_matrix = None
        self.confusion_matrix_colorscale = 'Viridis'

        self.prediction_figure = None
        self.roc_figure = None
        self.confusion_matrix_figure = None

        super(ClassifierPlotHandler, self).__init__(**params)

    @classmethod
    def from_pandas_dataframe(cls, dataframe, trained_classifier, **params):
        """
        Constructing the handler from a pandas dataframe.
        :param dataframe: the dataframe form which the handler is constructed.
        The 'target' column  should be included in the dataframe.
        :param trained_classifier: sklearn classifier (trained / fitted).
        :param params: other params.
        :return: returns the classifier plot handler object.
        """

        assert 'target' in dataframe.columns.values, 'target values not in dataframe'

        dataset = dict()
        dataset['data'] = dataframe.drop('target', axis=1).values
        dataset['target'] = dataframe['target'].values
        dataset['feature_names'] = dataframe.drop('target', axis=1).columns.values
        return cls(dataset, trained_classifier, **params)

    @property
    def trained_classifier(self):
        """
        The trained classifier .
        :return: The classifier in the sklearn format.
        """
        return self._trained_classifier

    @property
    def dataset(self):
        """
        The dataset
        :return: The dataset as a dictionary
        """
        return self._dataset

    @dataset.setter
    def dataset(self, dataset):
        """
        The dataset setter.
        :param dataset: the new dataset
        """
        self._dataset = dataset

    @property
    def predicted_target_score(self):
        """
        The predicted score - available if classifier has the predict_proba functionality.
        :return: The predicted score.
        """
        return self._predicted_target_score

    @property
    def confusion_matrix(self):
        """
        The confusion matrix.
        :return: The confusion matrix as a numpy array.
        """
        return self._confusion_matrix

    @property
    def n_classes(self):
        """
        The number of classes.
        :return:  An int representing the number of classes.
        """
        return self._n_classes

    def build_confusion_matrix(self, normalize=False):
        """
        Building the confusion matrix
        :param normalize: if True confusion matrix is normalized.
        """

        prediction = self.trained_classifier.predict(self._dataset['data'])

        self._confusion_matrix = confusion_matrix(self._dataset['target'], prediction)

        if normalize:
            self._confusion_matrix = \
                self._confusion_matrix.astype('float') / self._confusion_matrix.sum(axis=1)[:, np.newaxis]
        else:
            self._confusion_matrix = self._confusion_matrix

    def build_confusion_matrix_figure(self, figure_layout):
        """
        Builds the confusion matrix figure in confusion_matrix_figure.
        :param figure_layout: figure layout - plot.ly layout object.
        """

        if not self._confusion_matrix:
            self.build_confusion_matrix()

        cm = np.flipud(self._confusion_matrix)
        x = list(self.class_names)
        y = list(reversed(self.class_names))

        self.confusion_matrix_figure = ff.create_annotated_heatmap(z=cm, x=x, y=y,
                                                                   colorscale=self.confusion_matrix_colorscale)

        self.confusion_matrix_figure['layout'].update(figure_layout)

    def plot_confusion_matrix(self, figure_layout=None):
        """
        Plotting the confusion matrix figure with plot.ly's iplot function.
        :param figure_layout: figure layout - plot.ly layout object.
        """

        if not figure_layout:
            figure_layout = go.Layout(
                xaxis={'title': 'Confusion Matrix <br /><br />Predicted Value'},
                yaxis={'title': 'True Value'})

        if not self.confusion_matrix_figure:
            self.build_confusion_matrix_figure(figure_layout)
        else:
            self.confusion_matrix_figure['layout'].update(figure_layout)

        iplot(self.confusion_matrix_figure)

    def build_roc_figure(self, figure_layout=go.Layout()):
        """
        Building the ROC curve figure of the classifier.
        :param figure_layout: figure layout - plot.ly layout object.
        """

        data = list()

        if self.n_classes < 3:
            # False positive rate and true positive rate - computed from roc_curve()
            fpr, tpr, _ = roc_curve(self.dataset['target'], self.predicted_target_score[:, 1])

            # Area under curve.
            roc_auc = auc(fpr, tpr)

            # Updating the data list.
            data.append(go.Scatter(x=fpr,
                                   y=tpr,
                                   hoverinfo='y',
                                   mode='lines',
                                   line=dict(color='darkorange'),
                                   name='ROC curve (area = %0.2f)' % roc_auc))
        else:

            # False Positive, True Positive rates and Area Under Curve values for each class.
            fpr = dict()
            tpr = dict()
            roc_auc = dict()

            for i in range(self.n_classes):
                fpr[i], tpr[i], _ = roc_curve((self.dataset['target'] == i).astype(float),
                                              self.predicted_target_score[:, i])
                roc_auc[i] = auc(fpr[i], tpr[i])

                data.append(go.Scatter(x=fpr[i],
                                       y=tpr[i],
                                       hoverinfo='y',
                                       mode='lines',
                                       name='ROC curve of class {0} (area = {1:0.2f})'''.format(
                                           self.class_names[i], roc_auc[i])))

        # Diagonal
        data.append(go.Scatter(x=[0, 1], y=[0, 1],
                               mode='lines',
                               hoverinfo='skip',
                               line=dict(color='navy', dash='dash'),
                               showlegend=False))

        self.roc_figure = go.Figure(data=data, layout=figure_layout)

    def plot_roc(self, figure_layout=None):
        """
        Plotting the ROC curve figure with plot.ly's iplot function.
        :param figure_layout: figure layout - plot.ly Layout object.
        """

        if not figure_layout:
            figure_layout = go.Layout(title=dict(text='ROC Curve', x=0.5),
                                      xaxis=dict(title='False Positive Rate'),
                                      yaxis=dict(title='True Positive Rate'))

        if not self.roc_figure:
            self.build_roc_figure(figure_layout=figure_layout)
        else:
            self.roc_figure['layout'].update(figure_layout)

        iplot(self.roc_figure)

    def build_prediction_figure(self, figure_layout):
        """
        Building the classifier prediction figure.
        :param figure_layout: figure layout - plot.ly Layout object.
        """

        pass

    def plot_prediction(self, figure_layout=None):
        """
        Plotting the prediction figure with plot.ly's iplot function.
        :param figure_layout: figure layout - plot.ly Layout object.
        """

        if not figure_layout:
            figure_layout = go.Layout(title=dict(text='Classifier Prediction', x=0.5))

        if not self.prediction_figure:
            self.build_prediction_figure(figure_layout=figure_layout)
        else:
            self.prediction_figure['layout'].update(figure_layout)

        iplot(self.prediction_figure)

    def save_prediction_figure(self, file_name):
        """
        Saving the prediction figure as an html file.
        :param file_name: the html file name.
        """

        self.save_figure(self.prediction_figure, file_name)

    def save_roc_figure(self, file_name):
        """
        Saving the ROC curve figure as an html file.
        :param file_name: the html file name.
        """

        self.save_figure(self.roc_figure, file_name)

    def save_confusion_matrix_figure(self, file_name):
        """
        Saving the confusion matrix figure as an html file.
        :param file_name: the html file name.
        """

        self.save_figure(self.confusion_matrix_figure, file_name)


class TwoDimensionalClassifierPlotHandler(ClassifierPlotHandler):
    """ Handles all the plots related of the chosen classifier on 2D. """

    def __init__(self, dataset, trained_classifier, **params):
        """
        The initialization function of the 2D classifier plot handler.
        :param dataframe: the dataframe form which the handler is constructed.
        :param trained_classifier: sklearn classifier (trained / fitted).
        :param params: other params.
        """

        dataset['data'] = dataset['data'][:, :2]

        super(TwoDimensionalClassifierPlotHandler, self).__init__(dataset, trained_classifier, **params)

    def build_prediction_figure(self, figure_layout=go.Layout(), step_size=0.01):
        """
        Building the classifier prediction figure.
        :param figure_layout: figure layout - plot.ly Layout object.
        :param step_size: Plot resolution.
        """

        data = list()

        x_min, x_max = self.dataset['data'][:, 0].min() - 1, self.dataset['data'][:, 0].max() + 1
        y_min, y_max = self.dataset['data'][:, 1].min() - 1, self.dataset['data'][:, 1].max() + 1

        x = np.arange(x_min, x_max, step_size)
        y = np.arange(y_min, y_max, step_size)
        x_mesh, y_mesh = np.meshgrid(x, y)

        z = self.trained_classifier.predict(np.column_stack((x_mesh.ravel(), y_mesh.ravel())))

        z = z.reshape(x_mesh.shape)

        data.append(go.Contour(x=x, y=y, z=z,
                               showscale=False,
                               hoverinfo='skip',
                               colorscale='Viridis'))

        data.append(go.Scatter(x=self.dataset['data'][:, 0],
                               y=self.dataset['data'][:, 1],
                               text=[self.class_names[i] for i in self.dataset['target']],
                               hoverinfo='text',
                               mode='markers',
                               marker=dict(color=self.dataset['target'],
                                           showscale=False,
                                           colorscale='Reds',
                                           line=dict(color='black', width=1))))

        if 'feature_names' in self.dataset.keys():
            figure_layout['xaxis'].update({'title': self.dataset['feature_names'][0]})
            figure_layout['yaxis'].update({'title': self.dataset['feature_names'][1]})

        self.prediction_figure = go.Figure(data=data, layout=figure_layout)


class ThreeDimensionalClassifierPlotHandler(ClassifierPlotHandler):
    """ Handles all the plots related of the chosen classifier on 3D. """

    def __init__(self, dataset, trained_classifier, **params):
        """
        The initialization function of the 3D classifier plot handler.
        :param dataframe: the dataframe form which the handler is constructed.
        :param trained_classifier: sklearn classifier (trained / fitted).
        :param params: other params.
        """

        dataset['data'] = dataset['data'][:, :3]

        super(ThreeDimensionalClassifierPlotHandler, self).__init__(dataset, trained_classifier, **params)

    def build_prediction_figure(self, figure_layout=go.Layout()):
        """
        Plotting the classifier prediction and saving the figure.
        :param figure_layout: figure layout - plot.ly Layout object.
        """

        labels = self.trained_classifier.predict(self.dataset['data'])

        data = list()

        for label in set(labels):

            data_points = self.dataset['data'][np.in1d(labels, np.asarray(label))]

            data.append(go.Scatter3d(x=data_points[:, 0],
                                     y=data_points[:, 1],
                                     z=data_points[:, 2],
                                     text=self.class_names[label],
                                     hoverinfo='text',
                                     showlegend=True,
                                     name=self.class_names[label],
                                     mode='markers',
                                     marker=dict(
                                         line=dict(color='black', width=1))))

        if 'feature_names' in self.dataset.keys():
            figure_layout['scene'].update(
                dict(xaxis={'title': self.dataset['feature_names'][0]},
                     yaxis={'title': self.dataset['feature_names'][1]},
                     zaxis={'title': self.dataset['feature_names'][2]}))

        self.prediction_figure = go.Figure(data=data, layout=figure_layout)



import numpy as np

import plotly.graph_objs as go
from plotly.offline import iplot

from palantiri.BasePlotHandlers import PlotHandler


class RegressionPlotHandler(PlotHandler):
    """ Handles all the plots related of the chosen regressor. """

    def __init__(self, dataset, trained_regressor, **params):
        """
        Initialization function
        :param dataset: the dataset in a dict format with the following keys:
                        'data' - numpy array with all the data points.
                        'target' - the label of the corresponding data point.
                        'target_names' - the target name.

        :param trained_regressor: sklearn regressor(trained / fitted)..
        :param params: other params
        """

        self._dataset = dataset
        self._trained_regressor = trained_regressor

        self.prediction_figure = None

        self.target_name = 'Target'

        super(RegressionPlotHandler, self).__init__(**params)

    @classmethod
    def from_pandas_dataframe(cls, dataframe, trained_regressor, **params):
        """
        Constructing the handler from a pandas dataframe.
        :param dataframe: the dataframe form which the handler is constructed.
        :param trained_regressor: sklearn regressor (trained / fitted).
        :param params: other params.
        :return: returns the classifier plot handler object.
        """

        assert 'target' in dataframe.columns.values, 'target values not in dataframe'

        dataset = dict()
        dataset['data'] = dataframe.drop('target', axis=1).values
        dataset['target'] = dataframe['target'].values
        dataset['feature_names'] = dataframe.drop('target', axis=1).columns.values
        return cls(dataset, trained_regressor, **params)

    @property
    def trained_regressor(self):
        """
        The trained regressor.
        :return: The regressor in sklearn format.
        """
        return self._trained_regressor

    @property
    def dataset(self):
        """
        The dataset
        :return: The dataset as a dictionary
        """
        return self._dataset

    @dataset.setter
    def dataset(self, dataset):
        """
        The dataset setter.
        :param dataset: the new dataset
        """
        self._dataset = dataset

    def build_prediction_figure(self, figure_layout):
        """
        Building the regression figure.
        :param figure_layout: figure layout - plot.ly layout object.
        """

        pass

    def plot_prediction(self, figure_layout=None):
        """
        Plotting the regression figure with plot.ly's iplot function.
        :param figure_layout: figure layout - plot.ly layout object.
        """

        if not figure_layout:
            figure_layout = go.Layout(title=dict(text='Regression Plot', x=0.5))

        if not self.prediction_figure:
            self.build_prediction_figure(figure_layout)
        else:
            self.prediction_figure['layout'].update(figure_layout)

        iplot(self.prediction_figure)

    def save_prediction_figure(self, file_name):
        """
        Saving the prediction figure as an html file.
        :param file_name: the html file name.
        """

        self.save_figure(self.prediction_figure, file_name)


class OneDimensionalRegressionPlotHandler(RegressionPlotHandler):
    """ Handles all the plots related of the chosen 1D regression. """

    def __init__(self, dataset, trained_regressor, **params):
        """
        The initialization function of the 2D classifier plot handler.
        :param dataframe: the dataframe form which the handler is constructed.
        :param trained_regressor: sklearn regressor (trained / fitted).
        :param params: other params.
        """
        dataset['data'] = dataset['data'][:, :1]

        super(OneDimensionalRegressionPlotHandler, self).__init__(dataset, trained_regressor, **params)

    def build_prediction_figure(self, figure_layout, step_size=0.1, x_range=None):
        """
        Building the regression figure.
        :param figure_layout: figure layout - plot.ly layout object.
        :param step_size: resolution of the x-axis.
        :param x_range: the range of the prediction (x-axis),
        list of 2 numbers - indicating the start and end of the range
        if none will take the minimum and maximum of the data set.
        """

        if not x_range:
            x = np.arange(min(self.dataset['data']), max(self.dataset['data']), step_size).reshape(-1, 1)
        else:
            x = np.arange(min(x_range), max(x_range), step_size).reshape(-1, 1)

        data = [go.Scatter(x=self.dataset['data'][:, 0],
                           y=self.dataset['target'],
                           showlegend=False,
                           hoverinfo='skip',
                           mode='markers',
                           marker=dict(
                               line=dict(color='black', width=1))),
                go.Scatter(x=x.ravel(),
                           y=self.trained_regressor.predict(x).ravel(),
                           hoverinfo='y',
                           showlegend=False,
                           mode='lines')]

        if 'feature_names' in self.dataset.keys():
            figure_layout['xaxis'].update({'title': self.dataset['feature_names'][0]})
            figure_layout['yaxis'].update({'title': self.target_name})

        self.prediction_figure = go.Figure(data=data, layout=figure_layout)


class TwoDimensionalRegressionPlotHandler(RegressionPlotHandler):
    """ Handles all the plots related of the chosen regressor on 2D. """

    def __init__(self, dataset, trained_regressor, **params):
        """
        The initialization function of the 3D classifier plot handler.
        :param dataframe: the dataframe form which the handler is constructed.
        :param trained_regressor: sklearn regressor(trained / fitted).
        :param params: other params.
        """

        dataset['data'] = dataset['data'][:, :2]

        super(TwoDimensionalRegressionPlotHandler, self).__init__(dataset, trained_regressor, **params)

    def build_prediction_figure(self,  figure_layout=go.Layout(), x_range=None, y_range=None, step_size=0.1):
        """
        Building the regression figure.
        :param figure_layout: figure layout - plot.ly layout object.
        :param step_size: resolution of the x-axis.
        :param x_range: the range of the prediction (x-axis),
        list of 2 numbers - indicating the start and end of the range
        if none will take the minimum and maximum of the data set.
        :param y_range: similar to x_range for the y-axis.
        """

        if not x_range:
            x = np.arange(min(self.dataset['data'][:, 0]), max(self.dataset['data'][:, 0]), step_size)
        else:
            x = np.arange(min(x_range), max(x_range), step_size)

        if not y_range:
            y = np.arange(min(self.dataset['data'][:, 1]), max(self.dataset['data'][:, 1]), step_size)
        else:
            y = np.arange(min(y_range), max(y_range), step_size)

        x_mesh, y_mesh = np.meshgrid(x, y)

        z = self.trained_regressor.predict(np.column_stack((x_mesh.ravel(), y_mesh.ravel()))).reshape(x_mesh.shape)

        data = [go.Surface(x=x, y=y, z=z,
                           showscale=False,
                           colorscale='Viridis',
                           hoverinfo='z'),
                go.Scatter3d(x=self.dataset['data'][:, 0],
                             y=self.dataset['data'][:, 1],
                             z=self.dataset['target'],
                             hoverinfo='skip',
                             mode='markers',
                             marker=dict(showscale=False,
                                         colorscale='Reds',
                                         line=dict(color='black', width=0.3)))]

        if 'feature_names' in self.dataset.keys():
            figure_layout['scene'].update(
                dict(xaxis={'title': self.dataset['feature_names'][0]},
                     yaxis={'title': self.dataset['feature_names'][1]},
                     zaxis={'title': self.target_name}))

        self.prediction_figure = go.Figure(data=data, layout=figure_layout)



import numpy as np

import plotly.graph_objs as go
from plotly.offline import iplot

from palantiri.BasePlotHandlers import PlotHandler


class ClusteringPlotHandler(PlotHandler):
    """ Handles all the plots related of the chosen cluster. """

    def __init__(self, dataset, trained_cluster, **params):
        """
        Initialization function
        :param dataset: the dataset in a dict format with the following keys:
                        'data' - numpy array with all the data points.
        :param trained_cluster: sklearn cluster (trained / fitted).
        :param params: other params
        """

        self._dataset = dataset
        self._trained_cluster = trained_cluster

        self.prediction_figure = None

        if hasattr(self.trained_cluster, 'n_clusters'):
            self.n_clusters = self.trained_cluster.n_clusters
        elif hasattr(self.trained_cluster, 'labels_'):
            self.n_clusters = len([i for i in set(self.trained_cluster.labels_) if i >= 0])
        else:
            raise Exception('Number of clusters is not defined.')

        self.classes_names = ['Class {0}'.format(i) for i in range(self.n_clusters)]

        super(ClusteringPlotHandler, self).__init__(**params)

    @classmethod
    def from_pandas_dataframe(cls, dataframe, trained_cluster, **params):
        """
        Constructing the handler from a pandas dataframe.
        :param dataframe: the dataframe form which the handler is constructed.
        :param trained_cluster: sklearn cluster (trained / fitted).
        :param params: other params.
        :return: returns the classifier plot handler object.
        """

        dataset = dict()
        dataset['data'] = dataframe.drop('target', axis=1).values
        dataset['feature_names'] = dataframe.drop('target', axis=1).columns.values
        return cls(dataset, trained_cluster, **params)

    @property
    def trained_cluster(self):
        """
        The trained cluster.
        :return: The cluster in the sklearn format.
        """
        return self._trained_cluster

    @property
    def dataset(self):
        """
        The dataset
        :return: The dataset as a dictionary
        """
        return self._dataset

    @dataset.setter
    def dataset(self, dataset):
        """
        The dataset setter.
        :param dataset: the new dataset
        """
        self._dataset = dataset

    def build_prediction_figure(self, figure_layout):
        """
        Building the classifier prediction figure.
        :param figure_layout: figure layout - plot.ly Layout object.
        """

        pass

    def plot_prediction(self, figure_layout=None):
        """
        Plotting the prediction figure with plot.ly's iplot function.
        :param figure_layout: figure layout - plot.ly Layout object.
        """

        if not figure_layout:
            figure_layout = go.Layout(title=dict(text='Cluster Prediction', x=0.5))

        if not self.prediction_figure:
            self.build_prediction_figure(figure_layout=figure_layout)
        else:
            self.prediction_figure['layout'].update(figure_layout)

        iplot(self.prediction_figure)

    def save_prediction_figure(self, file_name):
        """
        Saving the prediction figure as an html file.
        :param file_name: the html file name.
        """

        self.save_figure(self.prediction_figure, file_name)


class TwoDimensionalClusteringPlotHandler(ClusteringPlotHandler):
    """ Handles all the plots related of the chosen cluster on 2D. """

    def __init__(self, dataset, trained_cluster, **params):
        """
        The initialization function of the 2D classifier plot handler.
        :param dataframe: the dataframe form which the handler is constructed.
        :param trained_cluster: sklearn cluster (trained / fitted).
        :param params: other params.
        """

        dataset['data'] = dataset['data'][:, :2]

        super(TwoDimensionalClusteringPlotHandler, self).__init__(dataset, trained_cluster, **params)

    def build_prediction_figure(self, figure_layout=go.Layout(), step_size=0.01):
        """
        Building the classifier prediction figure.
        :param figure_layout: figure layout - plot.ly Layout object.
        :param step_size: Plot resolution.
        """

        data = list()

        x_min, x_max = self.dataset['data'][:, 0].min() - 1, self.dataset['data'][:, 0].max() + 1
        y_min, y_max = self.dataset['data'][:, 1].min() - 1, self.dataset['data'][:, 1].max() + 1

        x = np.arange(x_min, x_max, step_size)
        y = np.arange(y_min, y_max, step_size)
        x_mesh, y_mesh = np.meshgrid(x, y)

        z = self.trained_cluster.predict(np.column_stack((x_mesh.ravel(), y_mesh.ravel())))

        z = z.reshape(x_mesh.shape)

        data.append(go.Contour(x=x, y=y, z=z,
                               showscale=False,
                               hoverinfo='skip',
                               colorscale='Viridis'))

        data.append(go.Scatter(x=self.dataset['data'][:, 0],
                               y=self.dataset['data'][:, 1],
                               text=[self.classes_names[i] for i in self.trained_cluster.predict(self.dataset['data'])],
                               hoverinfo='text',
                               mode='markers',
                               marker=dict(color=self.trained_cluster.predict(self.dataset['data']),
                                           showscale=False,
                                           colorscale='Reds',
                                           line=dict(color='black', width=1))))

        if 'feature_names' in self.dataset.keys():
            figure_layout['xaxis'].update({'title': self.dataset['feature_names'][0]})
            figure_layout['yaxis'].update({'title': self.dataset['feature_names'][1]})

        self.prediction_figure = go.Figure(data=data, layout=figure_layout)


class ThreeDimensionalClusteringPlotHandler(ClusteringPlotHandler):
    """ Handles all the plots related of the chosen cluster on 3D. """

    def __init__(self, dataset, trained_cluster, **params):
        """
        The initialization function of the 3D classifier plot handler.
        :param dataframe: the dataframe form which the handler is constructed.
        :param trained_cluster: sklearn cluster (trained / fitted).
        :param params: other params.
        """

        dataset['data'] = dataset['data'][:, :3]

        super(ThreeDimensionalClusteringPlotHandler, self).__init__(dataset, trained_cluster, **params)

    def build_prediction_figure(self, figure_layout=go.Layout()):
        """
        Plotting the cluster prediction and saving the figure.
        :param figure_layout: figure layout - plot.ly Layout object.
        """

        labels = self.trained_cluster.predict(self.dataset['data'])

        data = list()

        for label in set(labels):

            data_points = self.dataset['data'][np.in1d(labels, np.asarray(label))]
            data.append(go.Scatter3d(x=data_points[:, 0],
                                     y=data_points[:, 1],
                                     z=data_points[:, 2],
                                     text=self.classes_names[label],
                                     hoverinfo='text',
                                     showlegend=True,
                                     name=self.classes_names[label],
                                     mode='markers',
                                     marker=dict(
                                         line=dict(color='black', width=1))))

        if 'feature_names' in self.dataset.keys():
            figure_layout['scene'].update(
                dict(xaxis={'title': self.dataset['feature_names'][0]},
                     yaxis={'title': self.dataset['feature_names'][1]},
                     zaxis={'title': self.dataset['feature_names'][2]}))

        self.prediction_figure = go.Figure(data=data, layout=figure_layout)



from plotly import graph_objs as go
from plotly.offline import iplot
import colorlover as cl

from palantiri.BasePlotHandlers import PlotHandler


class RegretPlotHandler(PlotHandler):
    """
    The regret plot handler - handles all the regret related plots.
    """

    def __init__(self, cumulative_regret_data, **params):
        """
        Initialization function
        :param cumulative_regret_data: a dictionary where the keys are the dataset name and the values
               are T by N numpy arrays, where T is the horizon and N is the number of runs.
        :param params: other params
        """
        self._cumulative_regret_data = cumulative_regret_data
        self.regret_figure = None
        self.colors = cl.scales['7']['qual']['Accent']

        super(RegretPlotHandler, self).__init__(**params)

    def build_graph_figure(self, figure_layout=go.Layout()):
        """
        Building the regret plot figure.
        :param figure_layout: a plot.ly layout object.
        """
        assert len(self._cumulative_regret_data.keys()) <= 7, 'More Colors are needed, update the colors attribute.'
        data = list()
        colors = self.colors.copy()
        for dataset_name, dataset in self._cumulative_regret_data.items():
            color = colors.pop()
            mean = dataset.mean(axis=1)
            std = dataset.std(axis=1)

            x = list(range(len(mean)))
            x_reversed = x[::-1]
            y = mean.tolist()
            y_upper = (mean + std).tolist()
            y_lower = (mean - std).tolist()
            y_lower = y_lower[::-1]

            data.append(go.Scatter(
                x=x + x_reversed,
                y=y_upper + y_lower,
                fill='tozerox',
                fillcolor=color.replace('rgb', 'rgba').replace(')', ',0.2)'),
                line=dict(color='rgba(255,255,255,0)'),
                showlegend=False,
                name=dataset_name
            ))
            data.append(go.Scatter(
                x=x,
                y=y,
                line=dict(color=color),
                mode='lines',
                name=dataset_name,
            ))

        self.regret_figure = go.Figure(data=data, layout=figure_layout)

    def plot_regret(self, figure_layout=None):
        """
        Plotting the graph figure.
        :param figure_layout: a plot.ly layout object.
        """

        if not figure_layout:
            figure_layout = go.Layout(title=dict(text='Regret Plot', x=0.5),
                                      xaxis=dict(
                                          title='Round',
                                          autorange=True,
                                          showgrid=False,
                                          zeroline=False,
                                          showline=False,
                                          ticks='',
                                          showticklabels=True),
                                      yaxis=dict(
                                          title='Mean Accumulated Regret',
                                          autorange=True,
                                          showgrid=False,
                                          zeroline=False,
                                          showline=False,
                                          ticks='',
                                          showticklabels=True))

        if not self.regret_figure:
            self.build_graph_figure(figure_layout=figure_layout)

        iplot(self.regret_figure)



from plotly.offline import plot
import plotly.graph_objects as go
import matplotlib.pyplot as plt
import numpy as np


class PlotHandler(object):
    """ Base class for the plot handlers """

    def __init__(self, **params):
        """
        The initialization function.
        :param params: parameters.
        """

        for k, v in params.items():
            setattr(self, k, v)

    @staticmethod
    def save_figure(figure, file_name):
        """
        Saving the figure as an html file.
        :param figure: plot.ly figure graphical object to be saved.
        :param file_name: The html file name
        """

        with open(file_name, "w") as text_file:
            text_file.write('< script src = "https://cdn.plot.ly/plotly-latest.min.js" > < / script > \n')
            text_file.write(plot(figure, include_plotlyjs=False, output_type='div'))
